#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#define PI 3.14159265358979323846
#define MAX_ERROR 0.1

typedef struct { double re; double im; } complex;

complex *omega, *aux0, *aux1, *aux2, *aux3, *aux4;
double dfftbase, *daux;
unsigned int lfftbase, fftbase, lfftsize, fftsize, **ffte;

void
init_ffte()
{
  extern unsigned int lfftsize, fftsize, **ffte;
  unsigned int i, j, k;
  
  for ( i = 0 ; i < lfftsize ; i++ ) for ( j = 0 ; j < fftsize ; j++ )
  {
    ffte[i][j] = 0;
    for ( k = 0 ; k <= i ; k++ )
      if ( j & (1<<k) ) ffte[i][j] |= 1<<(lfftsize-1-k);
  }
}

void
init_omega()
{
  extern unsigned int fftsize;
  unsigned int i;

  for ( i = 0 ; i < fftsize ; i++ )
  { omega[i].re = cos((PI*2*i)/fftsize); omega[i].im = sin((PI*2*i)/fftsize); }
}

void
init_fft(unsigned int ls, unsigned int lb)
{
  extern unsigned int lfftsize, fftsize, lfftbase, fftbase, **ffte;
  extern complex *omega, *aux0, *aux1, *aux2, *aux3, *aux4;
  extern double dfftbase, *daux;
  unsigned int i;

  lfftsize = ls;
  fftsize = 1<<ls;
  lfftbase = lb;
  fftbase = 1<<lb;
  dfftbase = (double)(fftbase);

  ffte = (unsigned int **)(calloc(lfftsize,sizeof(int *)));
  for ( i = 0 ; i < lfftsize ; i++ )
    ffte[i] = (unsigned int *)(calloc(fftsize,sizeof(int)));
  omega = (complex *)(calloc(fftsize,sizeof(complex)));
  aux0 = (complex *)(calloc(fftsize,sizeof(complex)));
  aux1 = (complex *)(calloc(fftsize,sizeof(complex)));
  aux2 = (complex *)(calloc(fftsize,sizeof(complex)));
  aux3 = (complex *)(calloc(fftsize,sizeof(complex)));
  aux4 = (complex *)(calloc(fftsize,sizeof(complex)));
  daux = (double *)(calloc(fftsize,sizeof(double)));

  init_ffte();
  init_omega();
}

void
end_fft()
{
  extern unsigned int lfftsize, **ffte;
  extern complex *omega, *aux0, *aux1, *aux2, *aux3, *aux4;
  unsigned int i;
  
  for ( i = 0 ; i < lfftsize ; i++ ) free(ffte[i]);
  free(ffte);
  free(omega);
  free(aux0);
  free(aux1);
  free(aux2);
  free(aux3);
  free(aux4);
}

void
fft_step(unsigned int step, complex *pin, complex *pout)
{
  extern unsigned int fftsize, lfftsize, **ffte;
  extern complex *omega;

  unsigned int i, i0, i1, e;
  
  for ( i = 0 ; i < fftsize ; i++ )
  {
    i0 = i>>1;
    i1 = i0 | 1<<(lfftsize-1);
    e = ffte[step][i];
    pout[i].re = pin[i0].re + (omega[e].re * pin[i1].re) - (omega[e].im * pin[i1].im);
    pout[i].im = pin[i0].im + (omega[e].re * pin[i1].im) + (omega[e].im * pin[i1].re);
  }
}

void
fft(complex *pin, complex *pout)
{
  extern unsigned int lfftsize;
  extern complex *aux0, *aux1;
  unsigned int i;

  fft_step(0,pin,aux0);

  for ( i = 1 ; i < lfftsize ; i++ )
  { if ( i & 1 ) { fft_step(i,aux0,aux1); } else { fft_step(i,aux1,aux0); } }

  if ( lfftsize & 1 )
  {
    for ( i = 0 ; i < fftsize ; i++ )
    {
      pout[i].re = aux0[ffte[lfftsize-1][i]].re;
      pout[i].im = aux0[ffte[lfftsize-1][i]].im;
    }
  }
  else
  {
    for ( i = 0 ; i < fftsize ; i++ )
    {
      pout[i].re = aux1[ffte[lfftsize-1][i]].re;
      pout[i].im = aux1[ffte[lfftsize-1][i]].im;
    }
  }
}

void
ffti(complex *pin, complex *pout)
{
  extern unsigned int fftsize;
  extern complex *aux2;
  double is;
  unsigned int i;

  is = 1/((double)(fftsize));
  fft(pin,aux2);

  pout[0].re = is * aux2[0].re;
  pout[0].im = is * aux2[0].im;
  for ( i = 1 ; i < fftsize ; i++ )
  { 
    pout[i].re = is * aux2[fftsize-i].re;
    pout[i].im = is * aux2[fftsize-i].im;
  }
}

void
psquare(double *p)
{
  extern complex *aux3, *aux4;
  extern unsigned int fftsize;
  complex z;
  unsigned int i;

  for ( i = 0 ; i < fftsize ; i++ )
  { aux4[i].re = p[i]; aux4[i].im = 0.0; }
  fft(aux4,aux3);
  for ( i = 0 ; i < fftsize ; i++ )
  {
    z.re = aux3[i].re * aux3[i].re - aux3[i].im * aux3[i].im;
    z.im = 2.0 * aux3[i].re * aux3[i].im;
    aux3[i].re = z.re;
    aux3[i].im = z.im;
  }
  ffti(aux3,aux4);
  for ( i = 0 ; i < fftsize ; i++ ) p[i] = aux4[i].re;
}

unsigned int
isquare(int *n)
{
  unsigned int i, answ;
  extern double dfftbase, *daux;
  double x;

  answ = 0;

  for ( i = 0 ; i < fftsize ; i++ ) daux[i] = (double)(n[i]);
  psquare(daux);

  for ( i = 0 ; i < fftsize ; i++ )
  {
    x = rint(daux[i]/dfftbase);
    if ( x != 0.0 )
    {
      if ( i + 1 == fftsize )
      { answ |= 1; daux[i] -= x*dfftbase; } /* overflow */
      else
      { daux[i+1] += x; daux[i] -= x*dfftbase; }
    }
    
    x = rint(daux[i]);

    if ( fabs(x - daux[i]) > MAX_ERROR ) answ |= 2; /* roudoff danger */

    n[i] = (int)(x);
  }

  return(answ);
}

void
mmod(int *n, int p)
{
  int a, b, c, d, e, i, l, m;

  a = p/lfftbase;
  b = p%lfftbase;
  c = 1<<b;
  d = c - 1;
  e = 1<<(lfftbase-b);

  for ( i = a ; i < fftsize ; i++ )
  {
    m = n[i];
    n[i] = 0;
    l = m&d;
    n[i-a] += (m - l)/c;
    if ( i == a ) { n[i] += l; }
    else { n[i-a-1] +=  l*e; }
  }
}

